/* LISTS.CPP - Programmed by Maarten Hofman */
/* 200404: Started programming */
/* 290404: Add release functionality */
/* 040704: Fixed problem with transparancy */
/* 090704: Added animated textures */
/* 240704: Make sure we keep the parent if we unload a single blob */
/* 270904: Added the redo from start code */
/* 111004: Added the animation speed option */
/* 150205: Added load_all_sounds() */
/* 270205: Added setcopy variable sc to unloading blobs */
/* 060305: Added an additional style of object */
/* 090305: Added the U style (moving rivers) */
/* 100305: Removed debug statements */
/* 140305: Added update */
/* 090405: Added 'V' style */

#include "lists.h"
#include <string.h>

// void real_exit(int how);

CMs3dModel *models[MODEL_MAX];
blob *blobs[BLOB_MAX];
CTexture *texs[TEXTURE_MAX];
soundbyte *sounds[SOUND_MAX];
int mm,tm,bm,sm;

void redo_lists_from_start() {
int i;

  for(i=0;i<tm;i++) delete(texs[i]);
  for(i=0;i<mm;i++) delete(models[i]);
  for(i=0;i<bm;i++) delete(blobs[i]);
  bm=0;
  tm=0;
  mm=0;
}

blob *find_blob(char *nm) {
int i;

  for(i=0;i<bm;i++) {
    if(blobs[i]->has_name(nm)) return blobs[i];
  }
  return NULL;
}

/* Load animation template */
void load_animation(char *nm,char *fn,char *txtr);

void unload_blob(char *nm) {
int i,j;

  for(i=0;i<bm;i++) {
    if(blobs[i]->has_base_name(nm)) {
      delete blobs[i];
      for(j=i+1;j<bm;j++) {
       blobs[j-1]=blobs[j];
      }
      bm--;
      i--;
    }
  }
}

CTexture *load_texture_animation(char *buf,bool trans) {
FILE *fp;
char line[100],fn[100];
int size,speed;
CTexture *first,*current,*previous;
int i;

  fp=fopen(buf,"r");
  if(!fp) { 
    fprintf(stderr,"Couldn't open file!\n");
    return NULL;
  }
  fgets(line,100,fp);
  sscanf(line,"%d %d",&size,&speed);
  fgets(line,100,fp);
  line[strlen(line)-1]=0;
  sprintf(fn,"%s%s",RESOURCE_DIR,line);
  first=new CTexture();
  first->LoadImage(fn);
  first->set_name(buf);
  if(trans) first->set_trans();
  previous=first;
  for(i=0;i<size-1;i++) {
    fgets(line,100,fp);
    line[strlen(line)-1]=0;
    sprintf(fn,"%s%s",RESOURCE_DIR,line);
    fprintf(stderr,"Loading texture %s\n",fn);
    current=new CTexture();
    current->LoadImage(fn);
    current->set_name(buf);
    if(trans) current->set_trans();
    previous->set_next(current,speed);
    previous=current;
  }
  current->set_next(first,speed);
  fclose(fp);
  return first;
}

void load_model_and_texture(char *nm,char *mdl,char *txtr) {
int i,found;
int style;
char buf[100];

  style=0;
  for(i=0;i<bm;i++) {
    if(blobs[i]->has_name(nm)) {
      fprintf(stderr,"Blob %s already exists!\n",nm);
      return;
    }
  }
  fprintf(stderr,"Maxima: %d %d %d\n",tm,mm,bm);
  if(mdl[strlen(mdl)-1]=='i') {
    fprintf(stderr,"Detected animation!\n");
    load_animation(nm,mdl,txtr);
    return;
  }
  assert(bm<BLOB_MAX);
  blobs[bm]=new blob();
  blobs[bm]->set_name(nm);
  sprintf(buf,"%s%s",RESOURCE_DIR,mdl);
  found=0;
  for(i=0;i<mm;i++) {
    if(models[i]->has_name(buf)) {
      blobs[bm]->set_model(i);
      found=1;
    }
  }
  if(!found) {
    assert(mm<MODEL_MAX);
    models[mm]=new CMs3dModel();
    /* Files starting with T are transparant */
    if(mdl[0]=='T') {
      style=1;
      fprintf(stderr,"Model %s will be transparant.\n",mdl);
    }
    if(mdl[0]=='U') {
      style=5;
    }
    if(mdl[0]=='V') {
      style=9;
    }
    if(mdl[0]=='A') {
      style=2;
      fprintf(stderr,"Model %s will be afire.\n",mdl);
    } 
    if(!models[mm]->Load(buf,style)) {
      return;
    }
    blobs[bm]->set_model(mm);
    if(style&1) blobs[bm]->set_trans();
    mm++;
  }
  found=0;
  sprintf(buf,"%s%s",RESOURCE_DIR,txtr);
  for(i=0;i<tm;i++) {
    if(texs[i]->has_name(buf)) {
      blobs[bm]->set_texture(i);
      if(style&1) {
        texs[i]->set_trans();
        fprintf(stderr,"Texture %d marked transparant.\n",i);
      }
      found=1;
      break;
    }
  }
  if(!found) {
    assert(tm<TEXTURE_MAX);
    if(txtr[strlen(txtr)-1]=='3') {
      fprintf(stderr,"Texture animation detected!\n");
      texs[tm]=load_texture_animation(buf,(style&1)==1);
    }
    else {
      texs[tm]=new CTexture();
      texs[tm]->LoadImage(buf);
      if(style&1) {
        texs[tm]->set_trans();
        fprintf(stderr,"Texture %d marked transparant.\n",tm);
      }
    }
    blobs[bm]->set_texture(tm);
    tm++;
  }
  bm++;
}

void load_animation(char *nm,char *fn,char *txtr) {
FILE *fp;
char mdl[32],buf[100];
int parts,animtype,i,speed;
blob *parent,*current;
float *a,x,y,z;

  sprintf(buf,"%s%s",RESOURCE_DIR,fn);
  fp=fopen(buf,"r");
  if(!fp) { 
    fprintf(stderr,"Couldn't open file!\n");
    return;
  }
  fgets(buf,100,fp);
  sscanf(buf,"%s %d",mdl,&parts);
  load_model_and_texture(nm,mdl,txtr);
  parent=find_blob(nm);
  if(!parent) {
    fprintf(stderr,"Couldn't find parent blob!\n");
    return;
  }
  /* Ghost animation */
  if(parts==-1) {
    fprintf(stderr,"Using ghost animation!\n");
    parent->set_rotate(2);
    fclose(fp);
    return;
  }
  parent->set_rotate(1);
  for(i=0;i<parts;i++) {
    fgets(buf,100,fp);
    a=(float *)malloc(10*sizeof(float));
    if(sscanf(buf,"%s <%f,%f,%f> %d %f %f %f %f %f %f %f %f %f %f %d",mdl,&x,&y,
           &z,&animtype,&a[0],&a[1],&a[2],&a[3],&a[4],&a[5],&a[6],&a[7],
           &a[8],&a[9],&speed)!=16) {
      fprintf(stderr,"Uncompliant line: %s\n",buf);
    }
    if(!speed) {
      fprintf(stderr,"Line with no speed: %s\n",buf);
    }
    sprintf(buf,"%s%d",nm,i);
    fprintf(stderr,"Attempting to load %s as model %s with texture %s\n",buf,mdl,
           txtr);
    load_model_and_texture(buf,mdl,txtr);
    current=find_blob(buf);
    if(!current) {
      printf("Couldn't find blob!\n");
      return;
    }
    current->set_parent(parent);
    current->set_relative(x,y,z);
    current->set_animation_type(animtype,speed);
    current->set_animation(a);
  }
  fclose(fp);
}

void load_sound(char *how) {
int i;

  /* Check if the sound is already loaded */
  for(i=0;i<sm;i++) if(sounds[i]->has_name(how)) return;
  assert(sm<SOUND_MAX);
  sounds[sm]=new soundbyte();
  sounds[sm]->set_fragment_size(frag_size);
  sounds[sm]->Load(how);
  sm++;
}

void load_all_sounds() {
char buf[100];
FILE *fp;
char *loc;

  sprintf(buf,"%sindex.lst",AUDIO_DIR);
  printf("Loading audio index.lst\n");
  fp=fopen(buf,"r");
  if(!fp) {
    printf("Failed to open file!\n");
    return;
  }
  while(!feof(fp)) {  
    fgets(buf,99,fp);
    loc=strstr(buf,".");
    if(loc==NULL) continue;
    *loc=0;
    load_sound(buf);
  }
  fclose(fp);
}

void update_blob(char *nm,char *mdl,char *txtr,int ifif) {
blob *current;
blobcopy bc;
float cx,cy,cz;

  current=find_blob(nm);
  if(current) {
    if(ifif) {
      if(current->get_updated()) return;
    }
    current->set_copy(&bc);
    if(bc.parent==NULL) {
      cx=current->get_cx();
      cy=current->get_cy();
      cz=current->get_cz();
    }
    unload_blob(nm);
    load_model_and_texture(nm,mdl,txtr);
    current=find_blob(nm);
    if(current) {
      if(bc.parent!=NULL) {
        current->set_parent(bc.parent);
        current->set_relative(bc.ax,bc.ay,bc.az);
        current->set_animation_type(bc.animtype,bc.animspeed);
        current->set_animation(bc.animation);
        current->set_updated();
        current->set_angle(bc.angle);
        current->update();
      }
      else {
        current->move_coordinates(cx,cy,cz);
      }
    }
  }
}

int check_maxima() {
  if(mm>MODEL_MAX-3) return 1;
  if(tm>TEXTURE_MAX-3) return 1;
  if(bm>BLOB_MAX-3) return 1;
  if(tm<9) return 0;
  if(mm<90) return 0;
  if(bm<170) return 0;
  return 1;
}
